--> Library NickTag is a small library for share individual nicknames and avatars.

--> Basic Functions:
-- NickTag:SetNickname (name) -> set the player nick name, after set nicktag will broadcast the nick over addon guild channel.
-- 


local major, minor = "NickTag-1.0", 1
local NickTag, oldminor = LibStub:NewLibrary (major, minor)

if (not NickTag) then 
	return 
end

------------------------------------------------------------------------------------------------------------------------------------------------------
--> constants

	local CONST_INDEX_NICKNAME = 1
	local CONST_INDEX_COLOR = 2
	local CONST_INDEX_TEXTUREPATH = 3
	local CONST_INDEX_TEXTURECOORD = 4
	local CONST_INDEX_REVISION = 5

	local CONST_COMM_FULLPERSONA = 1
	local CONST_COMM_LOGONREVISION = 2
	local CONST_COMM_REQUESTPERSONA = 3
	
------------------------------------------------------------------------------------------------------------------------------------------------------
--> library stuff

	_G.NickTag = NickTag --> nicktag object over global container

	local pool --> pointer to the cache pool

	LibStub:GetLibrary ("AceComm-3.0"):Embed (NickTag)
	LibStub:GetLibrary ("AceSerializer-3.0"):Embed (NickTag)
	LibStub:GetLibrary ("AceTimer-3.0"):Embed (NickTag)
	local CallbackHandler = LibStub:GetLibrary ("CallbackHandler-1.0")
	NickTag.callbacks = NickTag.callbacks or CallbackHandler:New (NickTag)
	
	NickTag.embeds = NickTag.embeds or {}
	local embed_functions = {
		"SetNickname",
		"SetNicknameColor",
		"SetNicknameAvatar",
		"GetNickname",
		"GetNicknameAvatar",
		"GetNicknameColor",
		"GetNicknameTable"
	}
	function NickTag:Embed (target)
		for k, v in pairs (embed_functions) do
			target[v] = self[v]
		end
		self.embeds [target] = true
		return target
	end
	
	local enUS = LibStub("AceLocale-3.0"):NewLocale ("NickTag-1.0", "enUS", true)
	if (enUS) then
		enUS ["STRING_ERROR_1"] = "Your nickname is too long, max of 12 characters is allowed."
		enUS ["STRING_ERROR_2"] = "Only letters and two spaces are allowed."
		enUS ["STRING_ERROR_3"] = "You can't use the same letter three times consecutively, two spaces consecutively or more then two spaces."
	end
	
	local ptBR = LibStub("AceLocale-3.0"):NewLocale ("NickTag-1.0", "ptBR")
	if (ptBR) then
		ptBR ["STRING_ERROR_1"] = "Seu apelido esta muito longo, o maximo permitido sao 12 caracteres."
		ptBR ["STRING_ERROR_2"] = "Apenas letras, numeros e espacos sao permitidos no apelido."
		ptBR ["STRING_ERROR_3"] = "Voce nao pode usar a mesma letra mais de 2 vezes consecutivas, dois espacos consecutivos ou mais de 2 espacos."
	end
	
------------------------------------------------------------------------------------------------------------------------------------------------------
--> send and receive functions

	function NickTag:OnReceiveComm (_, data, _, source)

		local type, serial, arg3, name, realm =  select (2, NickTag:Deserialize (data))
		
		--> 0x1: received a full persona
		if (type == CONST_COMM_FULLPERSONA) then
			local receivedPersona = arg3
			
			if (serial ~= NickTag:GetSerial()) then

				local storedPersona = NickTag:GetNicknameTable (serial)
				if (not storedPersona) then
					storedPersona = NickTag:Create (serial)
				end
				
				if (storedPersona [CONST_INDEX_REVISION] < receivedPersona [CONST_INDEX_REVISION]) then
					storedPersona [CONST_INDEX_REVISION] = receivedPersona [CONST_INDEX_REVISION]
					storedPersona [CONST_INDEX_NICKNAME] = receivedPersona [CONST_INDEX_NICKNAME]
					storedPersona [CONST_INDEX_COLOR] = receivedPersona [CONST_INDEX_COLOR]
					storedPersona [CONST_INDEX_TEXTUREPATH] = receivedPersona [CONST_INDEX_TEXTUREPATH]
					storedPersona [CONST_INDEX_TEXTURECOORD] = receivedPersona [CONST_INDEX_TEXTURECOORD]
				end
			end
		
		--> 0x2: received a revision version from a guy which logon in the game
		elseif (type == CONST_COMM_LOGONREVISION) then
		
			local receivedRevision = arg3
			local storedPersona = NickTag:GetNicknameTable (serial)
			
			if (not storedPersona or storedPersona [CONST_INDEX_REVISION] < receivedRevision) then
				--> not sure how connected realms will work, but guess this will be fine
				if (realm ~= GetRealmName()) then
					name = name .. "-" .. realm
				end
				NickTag:ScheduleTimer ("RequestPersona", math.random (1, 20), name)
			end
			
		--> 0x3: someone requested my persona, so i need to send to him
		elseif (type == CONST_COMM_REQUESTPERSONA) then
			--> not sure how connected realms will work, but guess this will be fine
			if (realm ~= GetRealmName()) then
				name = name .. "-" .. realm
			end
			NickTag:SendPersona (name)
		end

	end

	NickTag:RegisterComm ("NickTag", "OnReceiveComm")

	--> after logon, we send our revision, who needs update my persona will send 0x3 (request persona) to me and i send back 0x1 (send persona)
	function NickTag:SendRevision()
		local battlegroup_serial = NickTag:GetSerial()
		if (not battlegroup_serial) then
			return
		end
		
		local myPersona = NickTag:GetNicknameTable (battlegroup_serial)
		if (myPersona) then
			NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_LOGONREVISION, battlegroup_serial, myPersona [CONST_INDEX_REVISION], UnitName ("player"), GetRealmName()), "GUILD")
		end
	end
	
	--> i received 0x2 and his persona is out of date here, so i need to send 0x3 to him and him will send 0x1.
	function NickTag:RequestPersona (target)
		NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_REQUESTPERSONA, 0, 0, UnitName ("player"), GetRealmName()), "WHISPER", target)
	end

	--> this broadcast my persona to entire guild when i update my persona or send my persona to someone who doesn't have it or need to update.
	function NickTag:SendPersona (target)
		local battlegroup_serial = NickTag:GetSerial()
		if (not battlegroup_serial) then
			return
		end
		
		if (target) then
			--> was requested
			NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_FULLPERSONA, battlegroup_serial, NickTag:GetNicknameTable (battlegroup_serial)), "WHISPER", target)
		else
			--> updating my own persona
			NickTag.send_scheduled = false
			--> need to increase 1 revision
			NickTag:IncRevision()
			--> broadcast over guild channel
			NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_FULLPERSONA, battlegroup_serial, NickTag:GetNicknameTable (battlegroup_serial)), "GUILD")
		end
	end

------------------------------------------------------------------------------------------------------------------------------------------------------
--> on logon stuff
	
	--> reset cache
	function NickTag:ResetCache()
		local guid = UnitGUID ("player")
		if (guid) then
			local player = NickTag:GetNicknameTable (guid)
			if (player) then
				local serial = NickTag:GetSerial (guid)
				for this_serial, _ in pairs (NickTag.pool_table) do
					if (this_serial ~= serial) then
						NickTag.pool_table [this_serial] = nil
					end
				end
				--vardump (NickTag.pool_table)
			else
				table.wipe (NickTag.pool_table)
			end
			
			NickTag.pool_table.nextrest = time() + (60*60*24*15) --> 15 days or 1296000 seconds
		end
	end
	
	function NickTag:OnEvent (event, arg1)
		if (event == "ADDON_LOADED") then
			if (arg1 == "!NickTag") then
				NickTag.pool_table = _G.NickTagCache or {nextrest = time() + (60*60*24*15)} --> main pool were personas are stored
				
				NickTag.pool_table.nextrest = NickTag.pool_table.nextrest or (time() + (60*60*24*15)) --> debug
				--NickTag.pool_table.nextrest = 1 --debug
				
				--> we need to cleanup our cache every two weeks.
				--> doing this, we avoid storing undesired data, like players which left the guild.
				--> so, our cache will always be clean and fresh.
				if (time() > NickTag.pool_table.nextrest) then
					NickTag:ScheduleTimer ("ResetCache", 10)
				end
				pool = NickTag.pool_table --> updating pointer
				
				NickTag:ScheduleTimer ("SendRevision", 30) --> after logon, wait few seconds before send my revision
			end
		elseif (event == "PLAYER_LOGOUT") then
			_G.NickTagCache = NickTag.pool_table or {} --> saving cache
		end
	end

	local frame = CreateFrame ("frame", nil, UIParent)
	frame:Hide()
	frame:RegisterEvent ("ADDON_LOADED");
	frame:RegisterEvent ("PLAYER_LOGOUT");
	frame:SetScript ("OnEvent", NickTag.OnEvent)
	
------------------------------------------------------------------------------------------------------------------------------------------------------
--> basic functions

	--> trim from from http://lua-users.org/wiki/StringTrim
	function trim (s)
		local from = s:match"^%s*()"
		return from > #s and "" or s:match(".*%S", from)
	end
	--
	local titlecase = function (first, rest)
		return first:upper()..rest:lower()
	end
	--
	local have_repeated = false
	local count_spaces = 0
	local check_repeated = function (char) 
		if (char == "  ") then
			have_repeated = true
		elseif (string.len (char) > 2) then
			have_repeated = true
		elseif (char == " ") then
			count_spaces = count_spaces + 1
		end
	end
	
	--> we need to keep game smooth checking and formating nicknames.
	--> SetNickname and names comming from other player need to be check.
	function NickTag:CheckName (name)
	
		--> as nicktag only work internally in the guild, we think that is not necessary a work filter to avoid people using bad language.
		
		name = trim (name)
		
		--> limit nickname to 12 characters, same as wow.
		local len = string.len (name)
		if (len > 12) then
			return false, LibStub ("AceLocale-3.0"):GetLocale ("NickTag-1.0")["STRING_ERROR_1"] --> error 1 = nickname is too long, max of 12 characters.
		end
		
		--> check if contain any non allowed characters, by now only accpet letters, numbers and spaces.
		--> by default wow do not accetp spaces, but here will allow.
		--> tested over lua 5.2 and this capture was okey with accents, not sure why inside wow this doesn't work.
		local notallow = string.find (name, "[^a-zA-Z%s]")
		if (notallow) then
			return false, LibStub ("AceLocale-3.0"):GetLocale ("NickTag-1.0")["STRING_ERROR_2"] --> error 2 = nickname only support letters, numbers and spaces.
		end
		
		--> check if there is sequencial repeated characters, like "Jasooon" were repeats 3 times the "o" character.
		--> got this from http://stackoverflow.com/questions/15608299/lua-pattern-matching-repeating-character
		have_repeated = false
		count_spaces = 0
		string.gsub (name, '.', '\0%0%0'):gsub ('(.)%z%1','%1'):gsub ('%z.([^%z]+)', check_repeated)
		if (count_spaces > 2) then
			have_repeated = true
		end
		if (have_repeated) then
			return false, LibStub ("AceLocale-3.0"):GetLocale ("NickTag-1.0")["STRING_ERROR_3"] --> error 3 = cant use the same letter three times consecutively, 2 spaces consecutively or 3 or more spaces.
		end
		
		return true
	end

	--> set the "player" nickname and schedule for send updated persona
	function NickTag:SetNickname (name)
		--> check data before
		assert (type (name) == "string", "NickTag 'SetNickname' expects a string on #1 argument.")
		
		--> check if the nickname is okey to allowed to use.
		local okey, errortype = NickTag:CheckName (name)
		if (not okey) then
			return false, errortype
		end
		
		--> here we format the text to match titles, e.g converts name like "JASON NICKSHOW" into "Jason Nickshow". 
		name = name:gsub ("(%a)([%w_']*)", titlecase)
		
		--> get player serial, note that serials are unique between battlegroups and we are using serial instead of full GUID just for reduce memory usage, 
		--> e.g guids are strings with 18 characters, serials are 8 digits number (or 9).
		local battlegroup_serial = NickTag:GetSerial()
		if (not battlegroup_serial) then
			return
		end
		
		--> get the full nick table.
		local nick_table = NickTag:GetNicknameTable (battlegroup_serial)
		if (not nick_table) then
			nick_table = NickTag:Create (battlegroup_serial, true)
		end
		
		--> change the nickname for the player nick table.
		nick_table [CONST_INDEX_NICKNAME] = name
		
		--> send the update for script which need it.
		NickTag.callbacks:Fire ("NickTag_Update", CONST_INDEX_NICKNAME)
		
		--> schedule a update for revision and broadcast full persona.
		--> this is a kind of protection for scripts which call SetNickname, SetColor and SetAvatar one after other, so scheduling here avoid three revisions upgrades and 3 broadcasts to the guild.
		if (not NickTag.send_scheduled) then
			NickTag.send_scheduled = true
			NickTag:ScheduleTimer ("SendPersona", 1)
		end
		
		return true
	end

	function NickTag:SetNicknameColor (r, g, b)
		if (type (r) == "table") then
			r, g, b = unpack (r)
		end
		
		--> check data before
		assert (r and g and b, "NickTag 'SetColor' bad color format. Usage NickTag:SetColor ({r, g, b} or r, g, b)")
		
		local battlegroup_serial = NickTag:GetSerial()
		if (not battlegroup_serial) then
			return
		end
		
		local nick_table = NickTag:GetNicknameTable (battlegroup_serial)
		if (not nick_table) then
			nick_table = NickTag:Create (battlegroup_serial, true)
		end
		
		nick_table [CONST_INDEX_COLOR][1] = r
		nick_table [CONST_INDEX_COLOR][2] = g
		nick_table [CONST_INDEX_COLOR][3] = b
		
		NickTag.callbacks:Fire ("NickTag_Update", CONST_INDEX_COLOR)
		
		if (not NickTag.send_scheduled) then
			NickTag.send_scheduled = true
			NickTag:ScheduleTimer ("SendPersona", 1)
		end
		
		return true
	end

	function NickTag:SetNicknameAvatar (texture, l, r, t, b)

		if (l == nil) then
			l, r, t, b = 0, 1, 0, 1
		elseif (type (l) == "table") then
			l, r, t, b = unpack (l)
		end
		
		--> check data before
		assert (texture and l and r and t and b, "NickTag 'SetAvatar' bad format. Usage NickTag:SetAvatar (texturepath [, L, R, T, B] or texturepath [, {L, R, T, B}])")
		
		local battlegroup_serial = NickTag:GetSerial()
		if (not battlegroup_serial) then
			return
		end
		
		local nick_table = NickTag:GetNicknameTable (battlegroup_serial)
		if (not nick_table) then
			nick_table = NickTag:Create (battlegroup_serial, true)
		end
		
		nick_table [CONST_INDEX_TEXTUREPATH] = texture
		nick_table [CONST_INDEX_TEXTURECOORD][1] = l
		nick_table [CONST_INDEX_TEXTURECOORD][2] = r
		nick_table [CONST_INDEX_TEXTURECOORD][3] = t
		nick_table [CONST_INDEX_TEXTURECOORD][4] = b
		
		NickTag.callbacks:Fire ("NickTag_Update", CONST_INDEX_TEXTUREPATH)
		
		if (not NickTag.send_scheduled) then
			NickTag.send_scheduled = true
			NickTag:ScheduleTimer ("SendPersona", 1)
		end
	end

	--> get the full nicktag table 
	function NickTag:GetNicknameTable (serial)
		--> check data before
		assert (serial, "NickTag 'Get' expects a number on #1 argument.")
		
		if (type (serial) == "string") then
			serial = NickTag:GetSerial (serial)
			if (not serial) then
				return
			end
		end
		
		return pool [serial]
	end

	--> return the avatar and the texcoord.
	function NickTag:GetNicknameAvatar (serial, default, silent)
		if (not silent) then
			assert (serial, "NickTag 'GetAvatar' expects a number or string on #1 argument.")
		end
		
		if (type (serial) == "string") then
			serial = NickTag:GetSerial (serial, silent)
		end
		
		if (serial) then
			local _table = pool [serial]
			if (not _table and default) then
				return default, {0, 1, 0, 1}
			elseif (not _table) then
				return "", {0, 1, 0, 1}
			end
			return _table [CONST_INDEX_TEXTUREPATH] or default "", _table [CONST_INDEX_TEXTURECOORD] or {0, 1, 0, 1}
		end
	end

	function NickTag:GetNicknameColor (serial, default, silent)
		if (not silent) then
			assert (serial, "NickTag 'GetNicknameColor' expects a number or string on #1 argument.")
		end
		
		if (type (serial) == "string") then
			serial = NickTag:GetSerial (serial, silent)
		end
		
		if (serial) then
			local _table = pool [serial]
			if (not _table) then
				return default or {1, 1, 1}
			end
			return _table [CONST_INDEX_COLOR] or default or {1, 1, 1}
		end
	end
	
	function NickTag:GetNickname (serial, default, silent)
		if (not silent) then
			assert (serial, "NickTag 'GetNickname' expects a number or string on #1 argument.")
		end
		
		if (type (serial) == "string") then
			serial = NickTag:GetSerial (serial, silent)
		end
		
		if (serial) then
			local _table = pool [serial]
			if (not _table) then
				return default or nil
			end
			return _table [CONST_INDEX_NICKNAME] or default or nil
		end
	end


------------------------------------------------------------------------------------------------------------------------------------------------------
--> internal functions
	
	--> create a empty nick table for the player
	function NickTag:Create (serial, isSelf)
		--> check data before
		assert (type (serial) == "number", "NickTag 'Create' expects a number on #1 argument.")
		
		--> check if alredy exists
		local alredy_have = pool [serial]
		if (alredy_have) then
			return alredy_have
		end
		
		--> create the table: name, color, texture path, texture texcoord, revision
		local n = { UnitName ("player"), false, false, false, 1 }
		
		--> if not my persona, set revision to 0, this make always get update after creation
		if (not isSelf) then
			n [CONST_INDEX_REVISION] = 0
		end
		
		pool [serial] = n
		return n
	end

	--> inc the revision of the player persona after update nick or avatar
	function NickTag:IncRevision()
		local battlegroup_serial = NickTag:GetSerial()
		if (not battlegroup_serial) then
			return
		end
		
		local nick_table = NickTag:GetNicknameTable (battlegroup_serial)
		if (not nick_table) then
			nick_table = NickTag:Create (battlegroup_serial, true)
		end
		
		nick_table [CONST_INDEX_REVISION] = nick_table [CONST_INDEX_REVISION] + 1
		
		return true
	end

	--> convert GUID into serial number
	function NickTag:GetSerial (serial, silent)
		if (not serial) then
			local guid = UnitGUID ("player")
			if (not guid) then
				return
			end
			serial = guid:sub (12, 18)
		else
			if (not silent) then
				assert (type (serial) == "string", "NickTag 'GetSerial' expects a GUID string on #1 parameter"..serial)
				assert (string.len (serial) > 17, "NickTag 'GetSerial' expects a GUID string on #1 parameter")
			else
				if (type (serial) ~= "string") then
					return
				elseif (string.len (serial) < 17) then
					return
				end
			end
			
			serial = serial:sub (12, 18)
		end
		return tonumber ("0x"..serial)
	end